# Copyright:   (c) Daryl Fox 2013
# Licence:     CC-BY
#              http://creativecommons.org/licenses/by/3.0/
#-------------------------------------------------------------------------------

import sys
import Queue
import numpy
from scipy import signal

def convolve_server(in_queue, out_queue, mask, scale=1):
    def empty_queue(queue):
        item = in_queue.get()  # Blocks
        if item is None:
            return None

        while not queue.empty():   # WARNING: Only works with single consumer
            try:
                item = in_queue.get_nowait()
            except Queue.Empty:
                break

            if item is None:
                return None

        return item

    #print "Started Server"
    processed = 0
    reach = 0
    try:
        while True:
            job_id, item, cycles = empty_queue(in_queue)
            #print "Starting job:", job_id
            if job_id is None:
                break

            field = (255 - item).astype(numpy.uint32)
            for _ in range(cycles):
                field = signal.convolve2d(field, mask, 'same', 'fill', 0) / scale
                field = field.clip(0, 255).astype(numpy.uint32)
            field_shifted = (255 - field) << 24

            rows = list(numpy.apply_along_axis(numpy.sum, 1, field_shifted))
            reach = rows.index(max(rows))

            #print "Finished job:", job_id
            out_queue.put((job_id, field_shifted, reach))
            del field
            processed += 1
        out_queue.put((None, processed, reach))
    except Exception as err:
        sys.stderr.write("ERROR: %s\n" % err.message)
        return processed

def test():
    import pygame
    import multiprocessing
    import random

    MAX_FPS = 160
    NEMESIS_SPEED = 0.1

    WINDOW_SIZE = 640, 480

    GOO_COLOR = (96, 96, 96, 192)
    #GOO_ARRAY = numpy.array(([1, 2, 1],
    #                         [2, 3, 2],
    #                         [1, 2, 1]), dtype=numpy.uint32)
    GOO_ARRAY = numpy.array(([0.125, 0.25, 0.125],
                             [0.25, 1.0, 0.25],
                             [0.125, 0.25, 0.125]), dtype=numpy.float)
    GOO_ARRAY /= 16.0
    #GOO_ARRAY = numpy.array(([0, 0, 1, 0, 0],
    #                         [0, 1, 2, 1, 0],
    #                         [1, 2, 3, 2, 1],
    #                         [0, 1, 2, 1, 0],
    #                         [0, 0, 1, 0, 0]), dtype=numpy.uint32)
    GOO_SCALE = numpy.max(GOO_ARRAY)

    pygame.init()

    # set the display mode
    winstyle = 0
    bestdepth = pygame.display.mode_ok(WINDOW_SIZE, winstyle, 32)
    screen = pygame.display.set_mode(WINDOW_SIZE, winstyle, bestdepth)

    background = pygame.Surface(WINDOW_SIZE, pygame.SRCALPHA)
    background.fill((0, 0, 0, 255))
    for i in range(8):
        x, y = 1, random.randrange(WINDOW_SIZE[1])
        background.set_at((x,y), (0, 0, 0, 0))
    #pygame.surfarray.pixels_alpha(background).astype(numpy.uint32)

    clock = pygame.time.Clock()

    growth = NEMESIS_SPEED
    frames = 0

    convolve_request = multiprocessing.Queue()
    convolve_results = multiprocessing.Queue()
    convolver = multiprocessing.Process(target=convolve_server,
                                        args=(convolve_request,
                                              convolve_results,
                                              GOO_ARRAY, GOO_SCALE))
    convolver.start()
    convolver_id = 0
    convolver_min = 0

    running = True
    while running:
        dt = clock.tick(MAX_FPS)
        update_goo = False
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                running = False
            elif event.type == pygame.KEYDOWN:
                if event.key == pygame.K_ESCAPE:
                    running = False
            elif event.type == pygame.MOUSEBUTTONDOWN:
                size = 8 * event.button
                pygame.draw.circle(background, (0, 0, 0, 255), event.pos, size)
                update_goo = True
        if not (dt and running):
            continue
        frames += 1

        # Calculate Movement
        delta = dt / 1000.0
        growth += delta
        if growth >= NEMESIS_SPEED:
            growth = 0.0
            field = pygame.surfarray.array_alpha(background)
            convolve_request.put((convolver_id, field, 1))
            convolver_id += 1
            update_goo = False
        elif update_goo:
            field = pygame.surfarray.array_alpha(background)
            convolve_request.put((convolver_id, field, 1))
            convolver_min = convolver_id
            convolver_id += 1
            update_goo = False


        try:
            job_id, array = convolve_results.get_nowait()
            while not convolve_results.empty():
                job_id, array = convolve_results.get_nowait()
            if job_id >= convolver_min:
                pygame.surfarray.blit_array(background, array)
        except Queue.Empty:
            pass

        fps = clock.get_fps()
        pygame.display.set_caption('Goo -- %.01f FPS' % fps)

        screen.fill((64, 64, 64, 255))
        screen.blit(background, (0, 0))
        pygame.display.flip()

    if convolver.is_alive():
        print "Sending shutdown to server"
        convolve_request.put((None, None))
        print "waiting for response"
        while True:
            job_id, convolves = convolve_results.get()
            if job_id is None:
                break
        print "Joining"
        convolver.join()
    else:
        print "Convolver crashed"

    print "%d convolves / %d frames" % (convolves, frames),
    print "= %.02f%%" % (float(convolves) / frames * 100.0)


if __name__ == '__main__':
    test()



