# Copyright:   (c) Daryl Fox 2013
# Licence:     CC-BY
#              http://creativecommons.org/licenses/by/3.0/
#-------------------------------------------------------------------------------

import pygame
import random

from collections import namedtuple # XXX


import config

import game_objects
import game_hud
import game_buttons
import game_blocks
import game_ui
import game_sounds

import util

def get_movement_keys():
    active = False
    direction = [0, 0]

    for ix, pressed in enumerate(pygame.key.get_pressed()):
        if pressed:
            active = True
            action = config.KEYS.get(ix)

            if action == 'UP':
                direction[1] = -config.PLAYER_SPEED if direction[1] == 0 else 0
            elif action == 'DOWN':
                direction[1] = config.PLAYER_SPEED if direction[1] == 0 else 0
            elif action == 'LEFT':
                direction[0] = -config.PLAYER_SPEED if direction[0] == 0 else 0
            elif action == 'RIGHT':
                direction[0] = config.PLAYER_SPEED if direction[0] == 0 else 0
    return direction, active

def image_sprite_images(filenames, size):
    images = []
    for filename in filenames:
        raw_image, raw_size = util.load_image(filename)
        images.append(pygame.transform.smoothscale(raw_image, size))
    return images

def circle_sprite_images(size, colors):
    images = []
    for color in colors:
        s = pygame.Surface(size, pygame.SRCALPHA)
        s.fill((0,0,0,0))
        pygame.draw.circle(s, color, (size[0]/2, size[1]/2), size[0]/2)
        images.append(s)
    return images

def generate_initial_obstruction(size, player, sprite_sizes):
    separation = list(player.size)
    start_x = player.position[0] + (3 * separation[0])
    return game_blocks.generate_field(size, start_x, separation, sprite_sizes)

def get_new_row_dist(player):
    return random.randrange(int(player.size[0] * 1.1), int(player.size[0] * 1.5))

def place_pickups(ranges, images, field_height):
    width, height = images[0].get_size()
    pickups = []

    for x, min_y, max_y in ranges:
        upper = max_y - height
        if upper - min_y > height:
            y = random.randrange(min_y, upper) + (height/2)
            p = game_objects.Pickup((x-width,y), images)
            p.update(0, 0)
            if p.rect.bottom < field_height:
                pickups.append(p)
    return pickups

def main(screen, effects, level=1):
    terrain = game_objects.Terrain(config.TILE_SIZE, config.SPOT_SIZE, config.FIELD_SIZE)
    goo = game_objects.Goo(config.GOO_ADVANCE_RATE, config.FIELD_SIZE)

    #######################################################################
    ## Build the sprites
    #######################################################################
    player_images = circle_sprite_images(config.PLAYER_SIZE, config.PLAYER_COLORS)
    player = game_objects.Player(player_images, config.FIELD_SIZE)
    player.position = config.PLAYER_START[:]
    player.update(0.0, (0,0))

    bomber_images = circle_sprite_images(config.BOMBER_SIZE, config.BOMBER_COLORS)
    poison_images = circle_sprite_images(config.POISON_SIZE, config.POISON_COLORS)
    pickup_images = circle_sprite_images(config.PICKUP_SIZE, config.PICKUP_COLORS)
    explosion_images = image_sprite_images(config.EXPLOSION_SEQUENCE,
                                           (config.EXPLOSION_RADIUS*2,
                                            config.EXPLOSION_RADIUS*2))

    obstruction_sprites, obstruction_sizes, Flare = game_blocks.load_obstructions()
    obs = pygame.sprite.RenderPlain()  # Obstruction list
    blockages, max_x, pickup_ranges = generate_initial_obstruction(config.FIELD_SIZE, player,
                                                                   obstruction_sizes)

    pickups = pygame.sprite.RenderPlain()
    pickups.add(place_pickups(pickup_ranges, pickup_images, config.FIELD_SIZE[1]))
    obs.add(blockages)
    new_row = get_new_row_dist(player) + max_x

    bombers = pygame.sprite.RenderPlain()  # Active Bombs
    notices = pygame.sprite.RenderPlain()  # Text Messages
    flare   = pygame.sprite.RenderPlain()  # Endgame Flare

    #######################################################################
    ## Build the interface elements
    #######################################################################
    field = pygame.Surface(config.FIELD_SIZE, pygame.SRCALPHA)
    hud = game_hud.Hud(config.HUD_SIZE, 'assets/banner.png', 'assets/cloth.png')

    #btn_images = [game_buttons.place_holder_button(BTN_SIZE, i) for i in range(BUTTONS)]
    btn_images = [pygame.image.load(b) for b in config.BUTTON_IMAGES]
    buttons = [game_buttons.Button(btn_images[i]) for i in range(config.BUTTONS)]
    box = game_buttons.ButtonBox(config.BUTTONS_SIZE, buttons,
                                 'assets/cloth.png', config.BUTTONS_POS)

    pause_overlay = game_ui.pause_overlay(config.WINDOW_SIZE)
    pause_screen = screen.copy()
    ghost_overlay = game_ui.ghost_overlay(config.WINDOW_SIZE)

    #######################################################################
    ## Initial State
    #######################################################################
    hud.followers = config.INITIAL_FOLLOWERS
    hud.distance = config.LEVEL_DISTANCES[level]

    damage_counter = 0.0
    goo_active = False
    paused = False
    win = False
    stats = dict(Eaten=0, Bomber=0, Turbo=0, Ghost=0, Poison=0,
                 Distance=0, Pickedup=0, Level=level)
    frames = 0  # Frame counter

    #######################################################################
    ## Start the sound
    #######################################################################
    effects['goo'].play(loops=-1)
    effects['goo'].set_volume(0.5)

    #######################################################################
    ## Start of the main game loop
    #######################################################################
    clock = pygame.time.Clock()
    update_hud = True
    state = 'Running'
    running = True

    while running:  # Main loop
        dt = clock.tick(config.MAX_FPS)
        power = None
        update_goo = False

        #######################################################################
        ## Empty the event queue
        #######################################################################
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                pygame.quit()
                import sys
                sys.exit()
            elif event.type == pygame.KEYDOWN:
                action = config.KEYS.get(event.key, '')
                ##if action == 'QUIT':
                ##    state = 'Quit'
                ##    running = False
                if action == 'PAUSE':
                    paused = not paused
                elif action == 'DEBUG':
                    util.save_screenshot(screen)
                elif action.startswith('POWER_'):
                    pwr = int(action[-1]) - 1
                    btn = box.buttons[pwr]
                    if btn.active and not btn.on_cooldown:
                        btn.activate(config.POWERS[pwr].cooldown)
                        hud.followers -= config.POWERS[pwr].cost
                        power = config.POWERS[pwr].name
                        stats[power] += config.POWERS[pwr].cost
                        update_hud = True
                    else:
                        effects['click'].play()
            elif event.type == pygame.ACTIVEEVENT:
                paused = not event.gain
            elif not paused:
                if (event.type == pygame.MOUSEBUTTONDOWN) and (event.button == 1):
                    action = box.click(event.pos)
                    if action is not None:
                        btn = box.buttons[action]
                        if btn.active and not btn.on_cooldown:
                            btn.activate(config.POWERS[action].cooldown)
                            hud.followers -= POWERS[action].cost
                            power = config.POWERS[action].name
                            stats[power] += config.POWERS[pwr].cost
                            update_hud = True
                        else:
                            effects['click'].play()
        if not (dt and running):
            continue
        elif paused:
            fps = clock.get_fps()
            pygame.display.set_caption('Avidus -- %.01f FPS -- PAUSED --' % fps)
            screen.blit(pause_screen, (0, 0))
            screen.blit(pause_overlay, (0, 0))
            pygame.display.flip()
            continue
        else:
            frames += 1
            dt_seconds = dt / 1000.0

        #######################################################################
        ## Player Movement Keys
        #######################################################################
        direction, active = get_movement_keys()
        goo_active |= active
        player.update(dt_seconds, direction)

        #######################################################################
        ## Power Activation
        #######################################################################
        if power == 'Turbo':
            player.powers['turbo'] = config.TURBO_TIME
            notices.add(game_ui.turbo_messages(config.FIELD_SIZE, config.TURBO_TIME,
                                               player.position))
            effects['turbo'].play(0, int(config.TURBO_TIME * 1000.0))
        elif power == 'Ghost':
            player.powers['ghost'] = config.GHOST_TIME
            notices.add(game_ui.ghost_banner(config.FIELD_SIZE, config.GHOST_TIME))
            notices.add(game_ui.ghost_messages(config.FIELD_SIZE, config.TURBO_TIME,
                                               player.position))
            effects['ghost'].play()
        elif power == 'Bomber':
            b = game_objects.Bomber(player.position,
                                    bomber_images, explosion_images,
                                    config.BOMB_DURATION, 'BOMB', 2)
            b.radius = config.EXPLOSION_RADIUS
            bombers.add(b)
            notices.add(game_ui.stay_messages(config.FIELD_SIZE, 3.0,
                                              player.position))
            effects['drop'].play()
        elif power == 'Poison':
            b = game_objects.Bomber(player.position,
                                    poison_images, poison_images,
                                    0.3, 'POISON', 4)
            b.radius = config.POISON_RADIUS
            bombers.add(b)
            notices.add(game_ui.stay_messages(config.FIELD_SIZE, 3.0, player.position))
            effects['drop'].play()

        #######################################################################
        ## Collilsion Detection: Obstacles
        #######################################################################
        hit = pygame.sprite.spritecollide(player, obs, False, pygame.sprite.collide_mask)
        if hit:
            if player.powers['ghost']:
                player.ghosting = True
            elif player.ghosting:
                edge = max([b.pos[0] + b.size[0] for b in hit]) + 2
                player.position[0] = edge
                player.last[0] = edge
            else:
                player.revert_move(player.bump)
        elif player.ghosting:
            player.ghosting = False

        #######################################################################
        ## Scroll world if player is too far right
        #######################################################################
        if player.bump:
            terrain.scroll(player.bump)
            player.scroll(player.bump)
            goo.scroll(-player.bump)
            hud.distance -= player.bump
            obs_move = player.bump
            update_hud = True
            player.bump = 0
        else:
            obs_move = 0

        #######################################################################
        ## Collilsion Detection: Villages (pickups)
        #######################################################################
        if not player.powers['ghost']:
            hit = pygame.sprite.spritecollide(player, pickups, True, pygame.sprite.collide_mask)
            for h in hit:
                gain = random.randint(*config.PICKUP_BONUS)
                stats['Pickedup'] += gain
                msg = '+ %d followers' % gain
                notices.add(game_ui.pickup_messages(msg, config.FIELD_SIZE,
                                                    config.TURBO_TIME, player.position))
                hud.followers += gain
                update_hud = True
                effects['pickup'].play()

        #######################################################################
        ## Shift scrolled object
        #######################################################################
        obs.update(dt_seconds, -obs_move)
        flare.update(dt_seconds, -obs_move)
        bombers.update(dt_seconds, obs_move)
        pickups.update(dt_seconds, obs_move)

        #######################################################################
        ## Generate new obstables
        #######################################################################
        new_row -= obs_move
        if new_row <= 0:
            if hud.distance > config.FIELD_SIZE[0]:
                blockages, max_x, p = game_blocks.generate_column(config.FIELD_SIZE[0],
                                                                  config.FIELD_SIZE[1],
                                                                  int(player.size[0]*1.5),
                                                                  obstruction_sizes)
                obs.add(blockages)
                new_row = get_new_row_dist(player) + max_x + 25
                pickups.add(place_pickups(p, pickup_images, config.FIELD_SIZE[1]))
            else:
                flares, max_x = game_ui.get_flares(config.FIELD_SIZE[0],
                                                   config.FIELD_SIZE[1], Flare)
                new_row = get_new_row_dist(player) / 2 + max_x
                flare.add(flares)

        #######################################################################
        ## Advance the GOO!
        #######################################################################
        if goo_active:
            goo.update(dt_seconds, update_goo)
            # Goo volume varies with goo.reach
            volume = min(1.0, (1.33 * goo.reach / float(config.FIELD_SIZE[0]) + 0.5))
            effects['goo'].set_volume(volume)

            # Damage player upon contact
            if goo.touches(player) and not player.powers['ghost']:
                damage_counter += dt_seconds * config.GOO_DAMAGE_RATE
                if damage_counter >= 1:
                    dmg = int(damage_counter)
                    eaten = min(hud.followers+1, dmg)
                    hud.followers -= eaten
                    stats['Eaten'] += eaten
                    damage_counter = max(0.0, damage_counter - dmg)
                    update_hud = True

            # Check to see if bombs have reached the goo
            for bomb in bombers.sprites():
                if bomb.exploding and bomb.effect in ('BOMB', 'BOMBLET'):  # The exploding kind.
                    goo.cut_blob(bomb.explode_position, bomb.radius)
                    if bomb.effect == 'BOMB':
                        if not bomb.expanded:
                            if (bomb.duration < config.BOMBLET_TIME):
                                bomb.expanded = True
                                bomb.radius = int(bomb.radius * 3)
                                bomblets = util.calc_bomblet_positions(config.EXPLOSION_RADIUS,
                                                                       bomb.explode_position)
                                for bomblet_pos in bomblets:
                                    bomblet = game_objects.Bomber(bomblet_pos,
                                                                  bomber_images, explosion_images,
                                                                  config.BOMB_DURATION,
                                                                  'BOMBLET', 2)
                                    bomblet.radius = config.EXPLOSION_RADIUS
                                    bomblet.explode()
                                    bombers.add(bomblet)
                elif goo.touches(bomb):  # The goo sets off bombs
                    bomb.explode()
                    bomb.rate = config.EXPLOSION_RATE
                    if bomb.effect == 'POISON':
                        goo.poisoned = config.POISON_DURATION
                        effects['poison'].play()
                        goo.cut_blob(bomb.explode_position, bomb.radius)
                    elif bomb.effect == 'BOMB':  # No sound for BOMBLETs
                        effects['bomb'].play()
                    if not bomb.spoken:
                        bomb.spoken = True
                        notices.add(game_ui.bomb_messages(config.FIELD_SIZE, 1.0,
                                                          bomb.position))

        #######################################################################
        ## Clean up after bombs
        #######################################################################
        if not bombers.sprites():
            effects['drop'].stop()

        #######################################################################
        ## Update HUD (Banner)
        #######################################################################
        if update_hud:
            update_hud = False
            hud.update()
            stats['Distance'] = config.LEVEL_DISTANCES[level] - hud.distance

            if hud.followers < 0:
                running = False
                win = False
                break
            elif hud.distance <= 0:
                running = False
                win = True
                break

        #######################################################################
        ## Update Buttons
        #######################################################################
        for ix, action in enumerate(config.POWERS):
            if action.cost > hud.followers:
                box.buttons[ix].active = False
            elif not box.buttons[ix].on_cooldown:
                box.buttons[ix].active = True
        box.update(dt_seconds)

        #######################################################################
        ## Draw All The Things!
        #######################################################################
        notices.update(dt_seconds)
        fps = clock.get_fps()
        pygame.display.set_caption('Avidus -- %.01f FPS' % fps)

        # Render the play field
        field.blit(terrain.image, (-terrain.position, 0))
        field.blit(player.image, player.rect)
        obs.draw(field)
        flare.draw(field)
        pickups.draw(field)
        bombers.draw(field)
        field.blit(goo.image, (0, 0))
        if player.powers['ghost']:
            field.blit(ghost_overlay, (0,0))
        notices.draw(field)

        # Render the screen
        screen.blit(field, config.FIELD_POS)
        screen.blit(hud.image, config.HUD_POS)
        screen.blit(box.image, config.BUTTONS_POS)
        pause_screen = screen.copy()

        # Update the screen
        pygame.display.flip()


    #######################################################################
    ## Stop the game
    #######################################################################
    effects['goo'].fadeout(1000)
    convolves = goo.stop()
##    print "Result:", state
##    print "%d convolves / %d frames" % (convolves, frames),
##    print "= %.02f%%" % (float(convolves) / frames * 100.0)
##    print stats
    return win, stats

if __name__ == '__main__':
    pygame.mixer.pre_init(22050, -16, 2, 1024)
    pygame.init()

    screen = setup_window(WINDOW_SIZE)
    effects = game_sounds.load_sounds()
    main(screen, effects)

    pygame.quit()






