# Copyright:   (c) Daryl Fox 2013
# Licence:     CC-BY
#              http://creativecommons.org/licenses/by/3.0/
#-------------------------------------------------------------------------------

import pygame
import util

class ButtonBox(object):
    background = (128, 51, 0)
    def __init__(self, size, buttons, overlay, base_pos=(0,0)):
        self.size = size
        self.buttons = buttons
        self.base_pos = base_pos

        self._box = pygame.Surface(size)
        self.overlay = pygame.image.load(overlay).subsurface(self._box.get_rect())
        self.targets = []
        self.update()

    @property
    def image(self):
        return self._box

    def update(self, dt_seconds=0.0):
        for button in self.buttons:
            button.update(dt_seconds)

        self._box.fill(self.background)
        pygame.draw.line(self._box, (212, 85, 0),
                         (0, 0), (0, self.size[1]), 1)
        self._box.blit(self.overlay, (0, 0))

        image_size = self.buttons[0].image.get_size()
        stack_size = image_size[1] * len(self.buttons)
        excess = (self.size[1] - stack_size - 0) / 2

        x = self.size[0] / 2 - image_size[0] / 2
        y_offset = excess / 2
        y_offset = 5  # XXX: Fix this.
        step = self.size[1] / len(self.buttons) - 1

        for button_ix, button in enumerate(self.buttons):
            y = button_ix * step + y_offset
            self._box.blit(button.image, (x, y))
            self.targets.append((x, y, x + image_size[0], y + image_size[1]))

    def click(self, click_pos):
        for ix, target in enumerate(self.targets):
            if target[0] <= (click_pos[0] - self.base_pos[0]) <= target[2]:
                if target[1] <= (click_pos[1] - self.base_pos[1]) <= target[3]:
                    #self.buttons[ix].action()
                    return ix

class Button(object):
    def __init__(self, image, active=True):
        self.active = active

        self.cooldown_time = 1.5
        self.cooldown = 0.0
        self.last_cool = 0.0

        self.src_img = image
        self.size = image.get_size()
        self.overlay = pygame.Surface(self.size, pygame.SRCALPHA)
        self.overlay.fill((255, 255, 255, 32))
        self.overlay_rect = self.overlay.get_rect()

        self._images = [pygame.Surface(self.size, pygame.SRCALPHA),
                        pygame.Surface(self.size, pygame.SRCALPHA)]
        self._images[0].blit(self.src_img, (0, 0))
        self._images[0].blit(self.overlay, (0, 0))
        self._images[1].blit(self.src_img, (0, 0))

    @property
    def image(self):
        return self._images[self.active]

    @property
    def on_cooldown(self):
        return self.cooldown > 0.0

    def activate(self, time=1.5):
        if self.active and not self.on_cooldown:
            self.cooldown = time
            self.cooldown_time = time

    def update(self, dt_seconds):
        self.last_cool = self.cooldown
        self.cooldown = max(0.0, self.cooldown - dt_seconds)

        if self.cooldown:
            covered = (self.cooldown / self.cooldown_time)
            self.overlay_rect.bottom = int(covered * self.size[1])

            self._images[1].blit(self.src_img, (0, 0))
            self._images[1].blit(self.overlay, self.overlay_rect)

        elif self.last_cool:
            self._images[1].blit(self.src_img, (0, 0))

def place_holder_button(size, number):
    assert 0 <= number <= 9

    font = util.find_font_by_height(size[1] / 2)

    button = pygame.Surface(size)
    button.fill((128, 51, 0))
    pygame.draw.rect(button, (212, 85, 0), button.get_rect(), 4)

    label = font.render(str(number), True, (255, 255, 255))
    label_rect = label.get_rect(center=button.get_rect().center)
    button.blit(label, label_rect)

    return button

if __name__ == '__main__':
    pygame.init()

    size = 60, 440
    buttons = 4
    image_size = size[0] - 8, size[1] / buttons - 6

    images = [place_holder_button(image_size, i) for i in range(buttons)]
    buttons = [Button(images[i]) for i in range(buttons)]

    box = ButtonBox(size, buttons, 'assets/cloth.png')

    def setup_window(window_size):
        winstyle = 0
        bestdepth = pygame.display.mode_ok(window_size, winstyle, 32)
        return pygame.display.set_mode(window_size, winstyle, bestdepth)

    screen = setup_window(size)

    clock = pygame.time.Clock()
    running = True
    while running:
        dt = clock.tick(30)
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                running = False
            elif event.type == pygame.KEYDOWN:
                if event.key == pygame.K_ESCAPE:
                    running = False
            elif event.type == pygame.MOUSEBUTTONDOWN:
                if event.button == 1:
                    b = box.click(event.pos)
                    #box.buttons[b].active = not box.buttons[b].active
                    box.buttons[b].activate()
        if not (dt and running):
            continue

        box.update(dt / 1000.0)
        screen.fill((0,0,0))
        screen.blit(box.image, (0, 0))

        pygame.display.flip()




