# Copyright:   (c) Daryl Fox 2013
# Licence:     CC-BY
#              http://creativecommons.org/licenses/by/3.0/
#-------------------------------------------------------------------------------

import pygame
import random

import config
import util
import instructions

class Faller(pygame.sprite.Sprite):
    def __init__(self, image, pos, speed, limit):
        super(Faller, self).__init__()
        self.image = image
        self.pos = list(pos)
        self.speed = speed
        self.limit = limit

        self.rect = self.image.get_rect(center=pos)
        self.movement = 0.0

    def update(self, dt_seconds):
        self.movement += self.speed * dt_seconds
        if self.movement >= 1.0:
            move = int(self.movement)
            self.pos[1] += move
            self.movement = max(0.0, self.movement - move)
        self.rect.center = self.pos

        if self.rect.top > self.limit:
            self.kill()

def generate_faller(image, top=True):
    size = image.get_size()

    x = random.randrange(config.WINDOW_SIZE[0])
    if top:
        y = -size[1]
    else:
        y = random.randrange(config.WINDOW_SIZE[1])
    speed = random.randint(80, 120)

    return Faller(image, (x,y), speed, config.WINDOW_SIZE[1] + size[1])


def start_menu(screen, effects):
    clock = pygame.time.Clock()

    banner = pygame.image.load('assets/banner.png')
    cloth = pygame.image.load('assets/cloth.png')
    cloth = cloth.convert_alpha()
    flare = pygame.image.load('assets/flare.png')
    flare = flare.convert_alpha()

    background = pygame.Surface(config.WINDOW_SIZE, pygame.SRCALPHA)
    background.fill((128, 51, 0))
    mid = screen.get_rect().center[0]

    font = util.find_font_by_height(32)
    text = [font.render('Welcome to Avidus:', True, (255, 255, 0)),
            font.render('Those Left Behind', True, (255, 255, 0)),
            font.render('[I] -- Instructions', True, (212, 85, 0)),
            font.render('[P] -- Play Game', True, (212, 85, 0)),
            font.render('[Q] -- Quit', True, (212, 85, 0)),
           ]
    heights = (110, 150, 230, 280, 330)

    menu = []
    for ix, txt in enumerate(text):
        rect = txt.get_rect(center=(mid, heights[ix]))
        menu.append((txt, rect))

    flare_target = 10
    flares = pygame.sprite.RenderPlain()
    initial = [generate_faller(flare, False) for _ in range(flare_target)]
    flares.add(initial)

    keys = {pygame.K_i: 'instructions',
            pygame.K_p: 'play',
            pygame.K_q: 'quit',
           }

    running = True
    while running:
        dt = clock.tick(config.MAX_FPS)
        action = None

        #######################################################################
        ## Empty the event queue
        #######################################################################
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                pygame.quit()
                import sys
                sys.exit()
            elif event.type == pygame.KEYDOWN:
                action = keys.get(event.key, None)

        #######################################################################
        ## Handle actions
        #######################################################################
        if action == 'instructions':
            effects['click'].play()
            instructions.show_instructions(screen, effects)
        elif action:
            effects['click'].play()
            running = False

        flares.update(dt / 1000.0)
        for i in range(flare_target - len(flares.sprites())):
            flares.add(generate_faller(flare))

        screen.blit(background, (0, 0))
        flares.draw(screen)
        screen.blit(banner, (0, 0))
        for item in menu:
            screen.blit(*item)
        screen.blit(cloth, (0, config.FIELD_POS[1]))

        fps = clock.get_fps()
        pygame.display.set_caption('Avidus -- %.01f FPS' % fps)
        pygame.display.flip()

    return action == 'play'

def level_results(screen, effects, title, stats):
    clock = pygame.time.Clock()
    background = screen.copy()

    cloth = pygame.image.load('assets/cloth.png')
    cloth = cloth.convert_alpha()
    background.blit(cloth, (0,0))

    mid = screen.get_rect().center[0]
    total = stats['Eaten'] + stats['Bomber'] + stats['Turbo'] + stats['Ghost'] + stats['Poison']

    font_b = util.find_font_by_height(34)
    font_s = util.find_font_by_height(26)
    text = [font_b.render(title, True, (255, 255, 0)),
            font_s.render('Distance Walked: %d' % stats['Distance'], True, (212, 85, 0)),
            font_s.render('Followers Gained: %d' % stats['Pickedup'], True, (212, 85, 0)),
            font_s.render('Lost for Bombing: %d' % stats['Bomber'], True, (212, 85, 0)),
            font_s.render('Lost for Turbo: %d' % stats['Turbo'], True, (212, 85, 0)),
            font_s.render('Lost for Ghosting: %d' % stats['Ghost'], True, (212, 85, 0)),
            font_s.render('Lost for Poisoning: %d' % stats['Poison'], True, (212, 85, 0)),
            font_s.render('Lost to Nemesis: %d' % stats['Eaten'], True, (212, 85, 0)),
            font_s.render('Total Killed: %d' % total, True, (212, 85, 0)),
           ]
    heights = [110] + range(170, 410, 30)

    prompt = font_s.render('Press any key to continue.', True, (255, 255, 0))
    prompt_pos = prompt.get_rect(midbottom=(mid, config.WINDOW_SIZE[1] - 10))

    for ix, txt in enumerate(text):
        rect = txt.get_rect(center=(mid, heights[ix]))
        background.blit(txt, rect)

    delay = 1.5
    running= True
    while running:
        dt = clock.tick(config.MAX_FPS)
        #######################################################################
        ## Empty the event queue
        #######################################################################
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                pygame.quit()
                import sys
                sys.exit()
            elif event.type in (pygame.KEYDOWN, pygame.MOUSEBUTTONDOWN):
                if delay < 0:
                    effects['click'].play()
                    running = False
        delay -= dt / 1000.0

        screen.blit(background, (0, 0))
        if delay < 0:
            screen.blit(prompt, prompt_pos)

        fps = clock.get_fps()
        pygame.display.set_caption('Avidus -- %.01f FPS' % fps)
        pygame.display.flip()

def main():
    stats = dict(Eaten=5, Bomber=6, Turbo=7, Ghost=8, Poison=120,
                 Distance=12000, Pickedup=23, Level=7)
    pygame.init()
    winstyle = 0
    bestdepth = pygame.display.mode_ok(config.WINDOW_SIZE, winstyle, 32)
    screen = pygame.display.set_mode(config.WINDOW_SIZE, winstyle, bestdepth)
    import game_sounds
    effects = game_sounds.load_sounds()
    level_win(screen, effects, stats)

if __name__ == '__main__':
    main()











