# Copyright:   (c) Daryl Fox 2013
# Licence:     CC-BY
#              http://creativecommons.org/licenses/by/3.0/
#-------------------------------------------------------------------------------

import pygame

import numpy
import multiprocessing
import Queue
import random

import tiler
import convolver

class Entity(pygame.sprite.Sprite):
    def __init__(self, position, images, limits, rate=5):
        pygame.sprite.Sprite.__init__(self)
        self.position = list(position)
        self.last = list(position)

        self._images = images

        self.size = images[0].get_size()
        self.border = [s / 2 for s in self.size]
        self.limits = [limits[i]+self.border[i] for i in range(2)]
        self.movement = [0.0, 0.0]

        self.image_ix = 0
        self.rate = rate
        self.frames = 0
        self.still_mover = False  # Animates when not moving

        self.rect = self._images[0].get_rect()
        self._masks = [pygame.mask.from_surface(i) for i in self._images]

        self.powers = dict(turbo=0.0, ghost=0.0)
        self.ghosting = False

    @property
    def image(self):
        return self._images[self.image_ix]
    @property
    def mask(self):
        return self._masks[self.image_ix]

    def update(self, dt_seconds, direction):
        if self.powers['turbo']:
            direction = [d*2 for d in direction]
        self.last = self.position[:]
        travel = [0, 0]
        for i in (0, 1):
            self.movement[i] += direction[i] * dt_seconds
            if abs(self.movement[i]) >= 1.0:
                travel[i] = int(self.movement[i])
                self.movement[i] = max(0.0, self.movement[i] - travel[i])
                self.position[i] = max(self.border[i], min(self.limits[i] - self.size[i],
                                                           self.position[i] + travel[i]))
        self.rect.center = self.position

        if self.still_mover or any(travel):
            self.frames += 1
            if self.frames >= self.rate:
                self.frames = 0
                self.image_ix = (self.image_ix + 1) % len(self._images)

        for power in self.powers:
            self.powers[power] = max(0.0, self.powers[power] - dt_seconds)

        return travel

    def scroll(self, dx):
        self.position[0] -= dx

    def revert_move(self, bump=0):
        self.position = [self.last[0] - bump, self.last[1]] #self.last[:]

class Player(Entity):
    def __init__(self, images, limits, rate=4):
        super(Player, self).__init__((0,0), images, limits, rate)
        #self.position = [self.size[0] * 2, limits[1] / 2 - self.size[1] / 2]

        self.x_limit = self.limits[0] / 2
        self.bump = 0
        self._fields = [pygame.surfarray.array_alpha(img).astype(numpy.bool) for img in self._images]

    @property
    def field(self):
        return self._fields[self.image_ix]

    def update(self, dt_seconds, direction):
        super(Player, self).update(dt_seconds, direction)
        self.bump = max(0, self.position[0] - self.x_limit)

class Pickup(Entity):
    def __init__(self, pos, images, rate=5):
        super(Pickup, self).__init__(pos, images, (9999, 9999), rate)
        self.still_mover = True

    def update(self, dt_seconds, dx):
        super(Pickup, self).update(dt_seconds, (0, 0))
        self.position[0] -= dx
        self.rect.center = self.position

        if self.rect.right <= 0:
            self.kill()

# This used to be just a souped up player... then: bloat.
class Bomber(Entity):
    def __init__(self, pos, images, exploding_images, duration, effect, rate=5):
        super(Bomber, self).__init__(pos, images, (9999, 9999), rate)
        self._fields = [pygame.surfarray.array_alpha(img).astype(numpy.bool) for img in self._images]
        self.exploding_images = exploding_images
        self.exploding_size = self.exploding_images[0].get_size()
        self.still_mover = True

        self.exploding = False
        self.duration = duration
        self.effect = effect
        self.explode_position = [0,0]
        self.expanded = False
        self.spoken = False

    @property
    def image(self):
        if self.exploding:
            return self.exploding_images[self.image_ix]
        else:
            return self._images[self.image_ix]
    @property
    def field(self):
        return self._fields[self.image_ix]

    def scroll(self, dx):
        super(Bomber, self).scroll(dx)
        self.explode_position[0] -= dx

    def explode(self):
        self.explode_position = self.position
        self.exploding = True
        self.rect = self.image.get_rect()

    def update(self, dt_seconds, dx):
        super(Bomber, self).update(dt_seconds, (0, 0))
        self.position[0] -= dx

        if self.exploding:
            self.duration -= dt_seconds
            if self.duration <= 0.0:
                self.kill()
        elif self.position[0] <= 0:
            self.kill()

class Goo(object):
    base_color = (96, 96, 96)
    phase_rate = 0.3
    array = numpy.array(([0.125, 0.25, 0.125],
                         [0.125,  1.0,  0.125],
                         [0.125, 0.25, 0.125]), dtype=numpy.float)
    cycles = 2
    reach_window = 5
    def __init__(self, rate, size, seeds=8):
        self._server_started = False
        self.rate = rate
        self.size = size

        self.growth = rate
        self.phase = 0.0

        self.seeth_rate = 0.8
        self.seeth_counter = 0.0

        self.color = list(self.base_color)
        self.color_goal = list(self.base_color)
        self.color_limits = [(int(c*0.8), min(255, int(c*1.2))) for c in self.color]

        self._reach = [0] * self.reach_window

        # Alpha channel for the goo
        self.mask = pygame.Surface(size, pygame.SRCALPHA)
        self.mask.fill((0, 0, 0, 255))

        # For fast convolution, use a low resolution version of the mask
        self.lo_res_mask = pygame.Surface((size[0]/2, size[1]/2), pygame.SRCALPHA)
        self.lo_res_size = self.lo_res_mask.get_size()

        # nemesis is the goo masked by mask.
        self.nemesis = pygame.Surface(size)  # NOTE: do not use SRCALPHA
        self.nemesis.set_colorkey((0, 0, 0))

        self._request = multiprocessing.Queue()
        self._results = multiprocessing.Queue()
        self._server = multiprocessing.Process(target=convolver.convolve_server,
                                               args=(self._request,
                                                     self._results,
                                                     self.array, 1))
        #self._server.daemon = True
        self._server.start()
        self._server_started = True

        # These are used to preform transactions with the convolve server
        self._convolver_id = 0
        self._convolver_min = 0

        # gone == no goo present on map
        self.gone = False
        self.poisoned = 0.0

    @property
    def reach(self):
        return sum(self._reach) / self.reach_window
    @reach.setter
    def reach(self, val):
        self._reach.append(val)
        self._reach.pop(0)

    @property
    def server_started(self):
        return self._server_started

    @property
    def image(self):
        return self.nemesis

    def seed_goo(self, seeds=8):
        self.mask = pygame.Surface(self.size, pygame.SRCALPHA)
        self.mask.fill((0, 0, 0, 255))
        for i in range(seeds):
            x, y = 1, random.randrange(self.size[1])
            self.add_blob((x,y))

    def add_blob(self, pos, radius=5):
        pygame.draw.circle(self.mask, (0, 0, 0, 0), pos, radius)

    def cut_blob(self, pos, radius=5):
        pygame.draw.circle(self.mask, (0, 0, 0, 255), pos, radius)

    def __del__(self):
        if self.server_started:
            self.stop()

    def stop(self):
        if self.server_started:
            convolves = 0
            if self._server.is_alive():
                #print "Sending shutdown to server...",
                self._request.put((None, None, 1))
                #print "Waiting for response."
                while True:
                    job_id, convolves, reach = self._results.get()
                    if job_id is None:
                        break
                #print "Halting server."
                self._server.join()
                self._server_started = False
            else:
                pass
                #print "Convolver crashed."
        return convolves

    def update(self, dt_seconds, force_update):
        if self.gone:
            self.seed_goo(8)

        if not self.poisoned:
            self.growth += dt_seconds

        if (self.growth >= self.rate) or force_update:
            self.growth = 0.0
            self.lo_res_mask = pygame.transform.scale(self.mask, self.lo_res_size)
            field = pygame.surfarray.array_alpha(self.lo_res_mask)
            self._request.put((self._convolver_id, field, self.cycles))
            if force_update:
                self._convolver_min = self._convolver_id
            self._convolver_id += 1

        try:
            job_id, array, reach = self._results.get_nowait()
            while not self._results.empty():
                job_id, array, reach = self._results.get_nowait()
            if job_id >= self._convolver_min:
                self.gone = array.min() > 0
                pygame.surfarray.blit_array(self.lo_res_mask, array)
                self.mask = pygame.transform.scale(self.lo_res_mask, self.size)
                self.reach = reach * 2
        except Queue.Empty:
            pass

        self.phase += dt_seconds
        while self.phase > self.phase_rate:
            self.phase = max(0.0, self.phase - self.phase_rate)
            if self.color == self.color_goal:
                self.color_goal = [random.randrange(*self.color_limits[i]) for i in range(3)]
            else:
                for rgb in range(3):
                    if self.color[rgb] > self.color_goal[rgb]:
                        self.color[rgb] -= 1
                    elif self.color[rgb] < self.color_goal[rgb]:
                        self.color[rgb] += 1
        self.poisoned = max(0.0, self.poisoned - dt_seconds)
        if self.poisoned:
            self.nemesis.fill((self.color[0]/2, self.color[1], self.color[2]/2))
            self.growth = 0.0
        else:
            self.nemesis.fill(self.color)

        self.nemesis.blit(self.mask, (0,0))

        self.seeth_counter += dt_seconds
        if self.seeth_counter > self.seeth_rate:
            self.seeth_counter = 0
            self.seeth_rate = random.random() * 0.5 + 0.3
            if self.reach:
                for i in range(5):
                    pos = random.randrange(self.reach), random.randrange(self.size[1])
                    self.cut_blob(pos, random.randrange(4, 12))

        return self.nemesis

    def touches(self, player):
        goo_field = pygame.surfarray.pixels_alpha(self.mask)
        base = max(0, player.rect.left), player.rect.top
        size = player.size

        region = goo_field[base[0]:base[0]+size[0], base[1]:base[1]+size[1]] < 255
        return numpy.any(region & player.field)

    def scroll(self, dx):
        self.mask.scroll(dx, 0)

class Terrain(object):
    colors = {'background': (0,   128, 0),
              'foreground': (192, 192, 0),
              'highlight':  (8,   192, 0)}
    def __init__(self, tile_size, spot_size, size):
        segment_size = size[0]/2, size[1]
        background_half = tiler.generate_map(segment_size, tile_size, spot_size, self.colors, 6)

        segment_width = background_half.get_size()[0]
        self.terrain = pygame.Surface((size[0]*2, size[1]), pygame.SRCALPHA)
        self.terrain.blit(background_half, (0, 0))
        self.terrain.blit(background_half, (segment_width, 0))
        self.terrain.blit(background_half, (segment_width*2, 0))
        self.terrain.blit(background_half, (segment_width*3, 0))  # Not needed

        self.position = 0
        self.width = size[0]

    @property
    def image(self):
        return self.terrain

    def scroll(self, dx):
        self.position = (self.position + dx) % self.width



