# Copyright:   (c) Daryl Fox 2013
# Licence:     CC-BY
#              http://creativecommons.org/licenses/by/3.0/
#-------------------------------------------------------------------------------

import pygame
import random

import util

TURBO_MESSAGES = ["You'll be faster without me.",
                  "Hurry, take my horse!",
                  "I'm just slowing you down.",
                  "Run! Leave me.",
                 ]

GHOST_MESSAGES = ["I know a way!",
                  "I'll scout the path.",
                  "Sneak through here.  I'll guard your back.",
                  "There's a way around."
                 ]

PICKUP_MESSAGES = ["Show us the way! %s",
                   "Help us! %s",
                   "Let's go! %s",
                   "We're doomed! %s",
                  ]

STAY_MESSAGES = ["Give me the bomb.",
                 "Remember me.",
                 "Watch my children.",
                 "There' only one way.",
                ]

DIE_MESSAGES = ["Got a present for ya!",
                "DIE! DIE! DIE!",
                "Ha-ha-ha-ha!",
                "This is it!",
               ]

class Mover(pygame.sprite.Sprite):
    def __init__(self, start_pos, target_pos, speed, images, image_rate, loiter_time):
        super(Mover, self).__init__()

        self.pos = list(start_pos)
        self.target = list(target_pos)
        self.loiter_time = loiter_time

        # This doesn't work
        dx = abs(start_pos[0] - target_pos[0])
        dy = abs(start_pos[1] - target_pos[1])
        if dx > dy:
            self.speed = (speed, speed * dy / dx)
        elif dx < dy:
            self.speed = (speed * dx / dy, speed)
        else:
            self.speed = (speed, speed)

        self.limits = [min(target_pos[0], start_pos[0]),
                       min(target_pos[1], start_pos[1]),
                       max(target_pos[0], start_pos[0]),
                       max(target_pos[1], start_pos[1])]

        self._images = images
        self._rects = [i.get_rect(center=start_pos) for i in images]

        self.image_rate = image_rate
        self._image_ix = 0

        self.movement = [0.0, 0.0]
        self.active = True
        self.frames = 0

        self.img_cycle = False

    @property
    def image(self):
        return self._images[self._image_ix]
    @property
    def rect(self):
        return self._rects[self._image_ix]
    @property
    def directions(self):  # This didn't suck originally
        d = []
        for i in xrange(2):
            if self.pos[i] < self.target[i]:
                d.append(1)
            elif self.pos[i] > self.target[i]:
                d.append(-1)
            else:
                d.append(0)
        return d

    def update(self, dt_seconds):
        self.frames += dt_seconds
        if self.frames >= self.image_rate:
            self.frames = 0.0
            if self.img_cycle:
                self._image_ix = (self._image_ix + 1) % len(self._images)
            else:
                self._image_ix = min(len(self._images) - 1, self._image_ix + 1)

        travel = [dt_seconds * self.speed[i] * self.directions[i] for i in range(2)]
        for i in (0, 1):
            self.movement[i] += travel[i]
            if abs(self.movement[i]) > 1.0:
                move = int(self.movement[i])
                self.pos[i] = max(self.limits[i], min(self.limits[i+2], self.pos[i] + move))
                self.movement[i] = max(0.0, self.movement[i] - move)
        self.rect.center = self.pos

        if self.pos == self.target:
            self.loiter_time -= dt_seconds
            self.active = self.loiter_time > 0.0

        if not self.active:
            self.kill()

def pause_overlay(size):
    font = util.find_font_by_width(int(size[0] * 0.8), 'PAUSED')
    txt = font.render('PAUSED', True, (0, 96, 255))

    cloth = pygame.image.load('assets/cloth.png')
    cloth = cloth.convert_alpha()

    pause = pygame.Surface(size, pygame.SRCALPHA)
    pause.fill((0,0,0,0))
    pause_rect = pause.get_rect()
    pause.blit(cloth, (0, 0))
    pause.blit(txt, txt.get_rect(center=pause_rect.center))

    return pause

def ghost_overlay(size):
    sand = pygame.image.load('assets/sand.png')
    sand = sand.convert_alpha()

    ghost = pygame.Surface(size, pygame.SRCALPHA)
    ghost.fill((0,0,0,0))
    ghost.blit(sand, (0, 0))

    return sand

def ghost_banner(size, duration):
    txt = 'GHOST MODE'
    font = util.find_font_by_width(size[0] / 2, txt)
    message = font.render(txt, True, (255, 255, 255, 198))

    start = size[0] / 2, 0
    target = size[0] / 2, size[1]
    speed = size[1] / float(duration)
    img_rate = duration
    loiter = 0.0

    return Mover(start, target, speed, [message], img_rate, loiter)

def turbo_messages(size, duration, pos, height=22):
    txt = random.choice(TURBO_MESSAGES)
    colors = [(c, 32, 32, c) for c in range(255, 0, -16)]
    return sacrifice_messages(txt, colors, size, duration, pos, height)

def ghost_messages(size, duration, pos, height=22):
    txt = random.choice(GHOST_MESSAGES)
    colors = [(c, c, c, c) for c in range(255, 128, -8)]
    return sacrifice_messages(txt, colors, size, duration, pos, height)

def pickup_messages(msg, size, duration, pos, height=22):
    txt = random.choice(PICKUP_MESSAGES) % msg
    colors = [(32, 32, c, c) for c in range(255, 0, -16)]
    return sacrifice_messages(txt, colors, size, duration, pos, height)

def stay_messages(size, duration, pos, height=22):
    txt = random.choice(STAY_MESSAGES)
    colors = [(c, c, c, c) for c in range(255, 128, -8)]
    return sacrifice_messages(txt, colors, size, duration, pos, height)

def bomb_messages(size, duration, pos, height=26):
    txt = random.choice(DIE_MESSAGES)
    font = util.find_font_by_height(height)
    img = font.render(txt, True, (64, 64, 64, 255))

    start = target = pos
    speed = 1
    img_rate = duration
    loiter = duration

    return Mover(start, target, speed, [img], img_rate, loiter)


def sacrifice_messages(txt, colors, size, duration, pos, height=22):
    font = util.find_font_by_height(height)

    images = []
    for color in colors:
        msg = font.render(txt, True, color)
        img = pygame.Surface(msg.get_size(), pygame.SRCALPHA)
        img.blit(msg, (0, 0))
        images.append(img)

    start = pos
    target = [1 - msg.get_size()[0],
              -height if pos[1] > (size[1] / 2) else size[1] + height]
    speed = 75
    img_rate = 0.6
    loiter = 0.0

    return Mover(start, target, speed, images, img_rate, loiter)

def get_flares(x, height, Flare):
    sep_range = Flare._size[1], int(Flare._size[1] * 1.5)
    dx_range = 0, Flare._size[0] / 2
    max_x = x + Flare._size[0]

    flares = []

    y = int(Flare._size[1] * -1.5)
    while y < height:
        y += random.randint(*sep_range)
        x_bump = random.randint(*dx_range)
        max_x = min(max_x, x_bump + Flare._size[0])
        flare = Flare((x + x_bump, y))
        flare.update(0.0, 0)
        flares.append(flare)

    return flares, max_x


def main():
    import random

    MAX_FPS = 60
    WINDOW_SIZE = 640, 480

    pygame.init()

    # set the display mode
    winstyle = 0
    bestdepth = pygame.display.mode_ok(WINDOW_SIZE, winstyle, 32)
    screen = pygame.display.set_mode(WINDOW_SIZE, winstyle, bestdepth)
    center = screen.get_rect().center
    clock = pygame.time.Clock()

    def add_banner():
        colors = [[random.randrange(255) for _ in range(3)] for _ in range(3)]
        pos = [random.choice((0, WINDOW_SIZE[i])) for i in range(2)]
        target = center
        speed = random.randrange(50, 250)
        img_rate = 0.5
        loiter = float(random.randrange(1, 5))
        font_size = random.randrange(12, 25)

        font = pygame.font.Font(pygame.font.get_default_font(), font_size)
        images = [font.render('Message -- %d -- %.1f' % (speed, loiter), True, c) for c in colors]

        return Mover(pos, target, speed, images, img_rate, loiter)

    banners = pygame.sprite.RenderPlain()
    ix = 0
    running = True
    while running:
        dt = clock.tick(MAX_FPS)
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                running = False
            elif event.type == pygame.KEYDOWN:
                if event.key == pygame.K_ESCAPE:
                    running = False
                elif event.key == pygame.K_SPACE:
                    banners.add(add_banner())
        if not (dt and running):
            continue

        banners.update(dt / 1000.0)

        screen.fill((128,128,128))
        banners.draw(screen)

        fps = clock.get_fps()
        pygame.display.set_caption('Banners -- %.01f FPS' % fps)
        pygame.display.flip()

if __name__ == '__main__':
    main()


