# Copyright:   (c) Daryl Fox 2013
# Licence:     CC-BY
#              http://creativecommons.org/licenses/by/3.0/
#-------------------------------------------------------------------------------

import pygame

from functools import partial
import random

def generate_edge_points(width, height, spot_size, spot_space):
    # Left/Right
    spots = height / spot_space
    ys = [random.randrange(spot_size, height-spot_size)  for i in range(spots)]
    left_right_spots = [(y, random.choice((0,2,3))) for y in ys]

    # Top/Bottom
    spots = width / spot_space
    xs = [random.randrange(spot_size, width-spot_size)  for i in range(spots)]
    top_bottom_spots = [(x, random.choice((1,2,3))) for x in xs]

    return left_right_spots, top_bottom_spots

def generate_spots(spot_size, foreground, highlight=None):
    if highlight is None:
        highlight = forground

    spot = pygame.Surface((spot_size, spot_size), pygame.SRCALPHA)
    spot.fill((0,0,0,0))  # Transparent
    rect_border = pygame.Rect((2, spot_size/4, spot_size-3, spot_size/2))
    rect_fore = pygame.Rect((3, spot_size/4-1, spot_size-4, spot_size/2-2))
    pygame.draw.ellipse(spot, highlight, rect_border)
    pygame.draw.ellipse(spot, foreground, rect_fore)
    spot_rect = spot.get_rect()

    spots = []

    # Type 0 (-)
    s = pygame.Surface((spot_size, spot_size), pygame.SRCALPHA)
    s.blit(spot, s.get_rect())
    spots.append(s)

    # Type 1 (|)
    s = pygame.Surface((spot_size, spot_size), pygame.SRCALPHA)
    s.blit(pygame.transform.rotate(spot, 90), s.get_rect())
    spots.append(s)

    # Type 2 (/)
    s = pygame.Surface((spot_size, spot_size), pygame.SRCALPHA)
    s_rot = pygame.transform.rotate(spot, -45)
    s_rot_rect = s_rot.get_rect(center=spot_rect.center)
    s.blit(s_rot, s_rot_rect)
    spots.append(s)

    # Type 3 (\)
    s = pygame.Surface((spot_size, spot_size), pygame.SRCALPHA)
    s_rot = pygame.transform.rotate(spot, 45)
    s_rot_rect = s_rot.get_rect(center=spot_rect.center)
    s.blit(s_rot, s_rot_rect)
    spots.append(s)

    return spots

def generate_tile(width, height, background, edge_spots, spot_sprites, num_spots):
    tile = pygame.Surface((width, height), pygame.SRCALPHA)
    tile.fill(background)

    # Left/Right edge
    for y, ix in edge_spots[0]:
        spot = spot_sprites[ix]
        left = spot.get_rect(center=(0, y))
        right = spot.get_rect(center=(width-1, y))
        tile.blit(spot, left)
        tile.blit(spot, right)

    # Top/Bottom edge
    for x, ix in edge_spots[1]:
        spot = spot_sprites[ix]
        top = spot.get_rect(center=(x, 0))
        bot = spot.get_rect(center=(x, height-1))
        tile.blit(spot, top)
        tile.blit(spot, bot)

    border = max(spot_sprites[0].get_size())

    for _ in range(num_spots):
        x = random.randrange(border, width-border-1)
        y = random.randrange(border, height-border-1)
        orientation = random.randrange(len(spot_sprites))

        spot = spot_sprites[orientation]
        rect = spot.get_rect(center=(x, y))
        tile.blit(spot, rect)

    return tile

def generate_map(screen_size, tile_size, spot_size, colors, mean_spots):
    #tile_size = tiles[0].get_size()
    spots = generate_spots(spot_size, colors['foreground'], colors['highlight'])
    edges = generate_edge_points(tile_size[0], tile_size[1], spot_size, spot_size*6)
    gen = partial(generate_tile, tile_size[0], tile_size[1], colors['background'], edges, spots)

    spot_range = range(mean_spots - (mean_spots/4), mean_spots + (mean_spots/4))
    tiles = [gen(random.choice(spot_range)) for _ in range(50)]

    terrain = pygame.Surface(screen_size, pygame.SRCALPHA)
    x, y = 0, 0
    while y < screen_size[1]:
        while x < screen_size[0]:
            terrain.blit(random.choice(tiles), (x,y))
            x += tile_size[0]
        x, y = 0, y + tile_size[1]
    return terrain

if __name__ == '__main__':
    pygame.init()

    window_size = 640, 480
    tile_size = 640/8, 480/8
    spot_size = 5

    scroll_speed = 100

    # set the display mode
    winstyle = 0
    bestdepth = pygame.display.mode_ok(window_size, winstyle, 32)
    screen = pygame.display.set_mode(window_size, winstyle, bestdepth)
    #bestdepth = pygame.display.mode_ok([i*2 for i in window_size], winstyle, 32)
    #screen = pygame.display.set_mode([i*2 for i in window_size], winstyle, bestdepth)

    spots = generate_spots(spot_size, (192, 192, 0), (8, 192, 0))
    edges = generate_edge_points(tile_size[0], tile_size[1], spot_size, 30)
    clock = pygame.time.Clock()

    colors = {'background': (0,   128, 0),
              'foreground': (192, 192, 0),
              'highlight':  (8,   192, 0)}
    map_size = window_size[0]/2, window_size[1]
    generate = partial(generate_map, map_size, tile_size, spot_size, colors, 6)

    background_half = generate()
    half_width = background_half.get_size()[0]
    background = pygame.Surface((window_size[0]*2, window_size[1]), pygame.SRCALPHA)
    #pygame.draw.rect(background_half, (255,0,0), background_half.get_rect(), 1)
    background.blit(background_half, (0, 0))
    background.blit(background_half, (half_width, 0))
    background.blit(background_half, (half_width*2, 0))

    distance = 0
    offset = 0.0
    direction = 0

    running = True
    while running:
        dt = clock.tick(30)
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                running = False
            elif event.type == pygame.KEYDOWN:
                if event.key == pygame.K_ESCAPE:
                    running = False
                elif event.key == pygame.K_SPACE:
                    background = generate()
        if not (dt and running):
            continue

        # Player Keys
        direction = 0
        for key, pressed in enumerate(pygame.key.get_pressed()):
            if pressed:
                if key == pygame.K_LEFT:
                    direction = -scroll_speed if direction == 0 else 0
                elif key == pygame.K_RIGHT:
                    direction = scroll_speed if direction == 0 else 0

        # Calculate Movement
        delta = dt / 1000.0
        offset += direction * delta
        travel = 0
        while offset >= 1.0:
            travel = int(offset)
            offset -= travel
        while offset <= -1.0:
            travel = int(offset)
            offset -= travel
        distance = (distance + travel) % (half_width * 1)

        fps = clock.get_fps()
        pygame.display.set_caption('Nemesis -- %d -- %.03f -- %.01f FPS' % (distance, offset, fps))

        screen.fill((0,0,0))
        screen.blit(background, (-distance, 0))
        pygame.display.flip()






